import logging
from pyfiglet import Figlet

def main():   

    # import importlib.metadata
    # package_name = __package__ or "openfund-core"
    # version = importlib.metadata.version("openfund-core")

    # 创建日志记录器并设置输出到屏幕
    logger = logging.getLogger(__name__)
    console_handler = logging.StreamHandler()
    logger.addHandler(console_handler)
    # # 设置日志级别为INFO
    logger.setLevel(logging.INFO)
    
    f = Figlet(font="standard")  # 字体可选（如 "block", "bubble"）
    logger.info(f"\n{f.renderText("OpenFund Core")}")
    


if __name__ == "__main__":
    main()



import logging
import time
import ccxt
import pandas as pd


from decimal import Decimal
from core.utils.OPTools import OPTools
from ccxt.base.exchange import ConstructorArgs


class Exchange:
    BUY_SIDE = 'buy'
    SELL_SIDE = 'sell'
    LONG_KEY = 'long'
    SHORT_KEY = 'short'
    SIDE_KEY = 'side'
    SYMBOL_KEY = 'symbol'
    ENTRY_PRICE_KEY = 'entryPrice'
    MARK_PRICE_KEY = 'markPrice'
    CONTRACTS_KEY = 'contracts'
    def __init__(self, config:ConstructorArgs, exchangeKey:str = "okx",) :
        # 配置交易所
        self.exchange = getattr(ccxt, exchangeKey)(config)
        self.logger = logging.getLogger(__name__)

    def getMarket(self, symbol:str):
        # 配置交易对
        self.exchange.load_markets()
        
        return self.exchange.market(symbol)
  
    def get_position_mode(self) -> str:
    
        try:
            # 假设获取账户持仓模式的 API
            response = self.exchange.private_get_account_config()
            data = response.get('data', [])
            if data and isinstance(data, list):
                # 取列表的第一个元素（假设它是一个字典），然后获取 'posMode'
                position_mode = data[0].get('posMode', 'single')  # 默认值为单向
                
                return position_mode
            else:
               
                return 'single'  # 返回默认值
        except Exception as e:        
            error_message = f"Error fetching position mode: {e}"
            self.logger.error(error_message)        
            raise Exception(error_message)  
        
    def get_tick_size(self,symbol) -> Decimal:     
        
        market = self.getMarket(symbol)
        if market and 'precision' in market and 'price' in market['precision']:            
            return OPTools.toDecimal(market['precision']['price'])
        else:
            raise ValueError(f"{symbol}: 无法从市场数据中获取价格精度")

    def amount_to_precision(self,symbol, contract_size):
        return self.exchange.amount_to_precision(symbol, contract_size)
       
    def set_leverage(self,symbol, leverage, mgnMode='isolated',posSide=None):
        try:
            # 设置杠杆
            params = {
                # 'instId': instId,
                'leverage': leverage,
                'marginMode': mgnMode
            }
            if posSide:
                params['side'] = posSide
                
            self.exchange.set_leverage(leverage, symbol=symbol, params=params)
            self.logger.info(f"{symbol} Successfully set leverage to {leverage}x")
        except Exception as e:
            error_message = f"{symbol} Error setting leverage: {e}"
            self.logger.error(error_message)        
            raise Exception(error_message)  
    # 获取价格精度
    def get_precision_length(self,symbol) -> int:
        tick_size = self.get_tick_size(symbol)
        return len(f"{tick_size:.15f}".rstrip('0').split('.')[1]) if '.' in f"{tick_size:.15f}" else 0

    def format_price(self, symbol, price:Decimal) -> str:
        precision = self.get_precision_length(symbol)
        return f"{price:.{precision}f}"
    
    def convert_contract(self, symbol, amount, price:Decimal, direction='cost_to_contract'):
        """
        进行合约与币的转换
        :param symbol: 交易对符号，如 'BTC/USDT:USDT'
        :param amount: 输入的数量，可以是合约数量或币的数量
        :param direction: 转换方向，'amount_to_contract' 表示从数量转换为合约，'cost_to_contract' 表示从金额转换为合约
        :return: 转换后的数量
        """

        # 获取合约规模
        market_contractSize = OPTools.toDecimal(self.getMarket(symbol)['contractSize'])
        amount = OPTools.toDecimal(amount)
        if direction == 'amount_to_contract':
            contract_size = amount / market_contractSize
        elif direction == 'cost_to_contract':
            contract_size = amount / price / market_contractSize
        else:
            raise Exception(f"{symbol} : {direction} 是无效的转换方向，请输入 'amount_to_contract' 或 'cost_to_contract'。")
        
        return self.amount_to_precision(symbol, contract_size)
    
    def close_position(self, symbol, position, params={}) -> dict:

        amount = abs(float(position['contracts']))

        if amount <= 0:
            self.logger.warning(f"{symbol}: position contracts must be greater than 0")
            return
            
        max_retries = 3
        retry_count = 0
        while retry_count < max_retries:
            
            try:
                side = position[self.SIDE_KEY]
                self.logger.debug(f"{symbol}: Preparing to close position, side= {side}, amount={amount}")
                position_mode = self.get_position_mode()  # 获取持仓模式
                if position_mode == 'long_short_mode':
                    # 在双向持仓模式下，指定平仓方向
                    # pos_side = 'long' if side == 'long' else 'short'
                    pos_side = side
                else:
                    # 在单向模式下，不指定方向
                    pos_side = 'net'
                orderSide = 'buy' if side == 'long' else 'sell'                
                
                td_mode = position['marginMode']
                params = {                    
                    'mgnMode': td_mode,
                    'posSide': pos_side,
                    # 当市价全平时，平仓单是否需要自动撤销,默认为false. false：不自动撤单 true：自动撤单
                    'autoCxl': 'true',
                    **params
            
                }

                # 发送平仓请求并获取返回值
                order = self.exchange.close_position(
                    symbol=symbol,
                    side=orderSide,
                    params=params
                )
                
                self.logger.info(f"{symbol} Close position response : {order}")
                return order
            
            except Exception as e:
                
                retry_count += 1
                if retry_count == max_retries:
                    error_message = f"{symbol} Error closing position : {str(e)}"
                    self.logger.error(error_message)
                    raise Exception(error_message)                
                else:
                    self.logger.warning(f"{symbol} 平仓失败，正在进行第{retry_count}次重试: {str(e)}")
                    time.sleep(0.1)  # 重试前等待0.1秒
                        
    def cancel_all_orders(self, symbol):
        max_retries = 3
        retry_count = 0
        
        while retry_count < max_retries:
            try:
                # 获取所有未完成订单
                params = {
                    # 'instId': instId
                }
                open_orders = self.exchange.fetch_open_orders(symbol=symbol, params=params)
                
                # 批量取消所有订单
                if open_orders:
                    order_ids = [order['id'] for order in open_orders]
                    self.exchange.cancel_orders(order_ids, symbol, params=params)
                    
                    self.logger.debug(f"{symbol}: {order_ids} 挂单取消成功.")
                else:
                    self.logger.debug(f"{symbol}: 无挂单.")
                return True
                
            except Exception as e:
                retry_count += 1
                if retry_count == max_retries:
                    error_message = f"{symbol} 取消挂单失败(重试{retry_count}次): {str(e)}"
                    self.logger.error(error_message)
                    raise Exception(error_message)                
                else:
                    self.logger.warning(f"{symbol} 取消挂单失败，正在进行第{retry_count}次重试: {str(e)}")
                    time.sleep(0.1)  # 重试前等待0.1秒

    def cancel_all_algo_orders(self, symbol, attachType=None) -> bool:
        """_summary_

        Args:
            symbol (_type_): _description_
            attachType (_type_, optional): "TP"|"SL". Defaults to None.
        """
        
        params = {
            "ordType": "conditional",
        }
        try:
            orders = self.fetch_open_orders(symbol=symbol,params=params)
        except Exception as e:
            error_message = f"!!{symbol} : Error fetching open orders: {e}"
            self.logger.error(error_message)
            raise Exception(error_message)
       
        
        if len(orders) == 0:           
            self.logger.debug(f"{symbol} 未设置策略订单列表。")
            return True
     
        algo_ids = []
        if attachType and attachType == 'SL':
            algo_ids = [order['id'] for order in orders if order['stopLossPrice'] and order['stopLossPrice'] > 0.0 ]
        elif attachType and attachType == 'TP':
            algo_ids = [order['id'] for order in orders if order['takeProfitPrice'] and order['takeProfitPrice'] > 0.0]
        else :
            algo_ids = [order['id'] for order in orders ]
            
        if len(algo_ids) == 0 :         
            self.logger.debug(f"{symbol} 未设置策略订单列表。")
            return True
        
        max_retries = 3
        retry_count = 0
        
        while retry_count < max_retries:
            try:
                params = {
                    "algoId": algo_ids,
                    "trigger": 'trigger'
                }
                rs = self.exchange.cancel_orders(ids=algo_ids, symbol=symbol, params=params)
                
                return len(rs) > 0
       
            except Exception as e:
                retry_count += 1
                if retry_count == max_retries:
                    error_message = f"!!{symbol} : Error cancelling order {algo_ids}: {e}"
                    self.logger.error(error_message)
                    raise Exception(error_message) 
                   
                self.logger.warning(f"{symbol} : Error cancelling order {algo_ids}: {str(e)}")
                time.sleep(0.1)  # 重试前等待0.1秒 
    
    def place_algo_orders(self, symbol, position, price: Decimal, order_type, sl_or_tp='SL', params={}) -> bool:
        """
        下单
        Args:
            symbol: 交易对
            position: 仓位
            price: 下单价格
            order_type: 订单类型
        """
        # 计算下单数量
        amount = abs(position[self.CONTRACTS_KEY])
    
        if amount <= 0:
            self.logger.warning(f"{symbol}: amount is 0 for {symbol}")
            return

        # 止损单逻辑 
        adjusted_price = self.format_price(symbol, price)
        
        # 默认市价止损，委托价格为-1时，执行市价止损。
        sl_params = {
            **params,
            'slTriggerPx':adjusted_price , 
            'slOrdPx':'-1', # 委托价格为-1时，执行市价止损
            # 'slOrdPx' : adjusted_price,
            'slTriggerPxType':'last',
            'tdMode':position['marginMode'],
            'sz': str(amount),        
            'cxlOnClosePos': True,
            'reduceOnly':True,            
        }
        
        tp_params = {  
            **params,
            'tpTriggerPx':adjusted_price,
            'tpOrdPx' : adjusted_price,
            'tpOrdKind': 'condition',
            'tpTriggerPxType':'last',            
            'tdMode':position['marginMode'],
            'sz': str(amount),          
            'cxlOnClosePos': True,
            'reduceOnly':True
        }
        
        order_params = sl_params if sl_or_tp == 'SL' else tp_params
        # order_params.update(params)
        
        if order_type == 'limit' and sl_or_tp =='SL':
            order_params['slOrdPx'] = adjusted_price
        
        orderSide = self.BUY_SIDE if position[self.SIDE_KEY] == self.SHORT_KEY else self.SELL_SIDE # 和持仓反向相反下单
        
        order = {
            'symbol': symbol,
            'side': orderSide,
            'type': order_type,
            'amount': amount,
            'price': adjusted_price,
            'params': order_params
        }
        
        max_retries = 3
        retry_count = 0
        self.logger.info(f"{symbol} : Pre Algo Order placed:  {order} ")
        while retry_count < max_retries:
            try:
                # 创建订单
                order_result = self.exchange.create_order(**order)
                self.logger.info(f"{symbol} : --------- ++ Algo Order placed done. --------")
                return True
                
            except (ccxt.NetworkError, ccxt.ExchangeError, Exception) as e:
                retry_count += 1
                error_type = "网络" if isinstance(e, ccxt.NetworkError) else "交易所" if isinstance(e, ccxt.ExchangeError) else "未知"
                self.logger.warning(f"{symbol} : 设置止盈止损单时发生{error_type}错误,正在进行第{retry_count}次重试: {str(e)}")
                
                if retry_count == max_retries:
                    error_message = f"!! {symbol}: 设置止盈止损单失败(重试{max_retries}次): {str(e)}"
                    self.logger.error(error_message)
                    raise Exception(error_message)
                    
                time.sleep(0.1)
        
    def place_order(self, symbol, price: Decimal, amount_usdt, side, leverage=20, order_type='limit', params={}) -> bool: 
        """
        下单
        Args:
            symbol: 交易对
            price: 下单价格
            amount_usdt: 下单金额
            side: 下单方向
            order_type: 订单类型
        """       
        # 格式化价格
        adjusted_price = self.format_price(symbol, price)

        if amount_usdt <= 0:
            self.logger.warning(f"{symbol}: amount_usdt must be greater than 0")
            return
    
        pos_side = self.LONG_KEY if side == self.BUY_SIDE else self.SHORT_KEY

        # 设置杠杆 
        self.set_leverage(symbol=symbol, leverage=leverage, mgnMode='isolated',posSide=pos_side)  
        # 20250220 SWAP类型计算合约数量 
        contract_size = self.convert_contract(symbol=symbol, price = OPTools.toDecimal(adjusted_price) ,amount=amount_usdt)

        order_params = {
            **params,
            "tdMode": 'isolated',
            "side": side,
            "ordType": order_type,
            "sz": contract_size,
            "px": adjusted_price
        } 
        
        # # 模拟盘(demo_trading)需要 posSide
        # if self.is_demo_trading == 1 :
        #     params["posSide"] = pos_side
            
        order = {
            'symbol': symbol,
            'side': side,
            'type': order_type,
            'amount': contract_size,
            'price': adjusted_price,
            'params': order_params
        }
        
        max_retries = 3
        retry_count = 0

        self.logger.info(f"{symbol} : Pre Order placed:  {order} ")
        
        while retry_count < max_retries:
            try:
                # 使用ccxt创建订单
                order_result = self.exchange.create_order(**order)
                self.logger.info(f"{symbol} : --------- ++ Order placed done. --------")  
                return True
            except (ccxt.NetworkError, ccxt.ExchangeError, Exception) as e:
                retry_count += 1
                error_type = "网络" if isinstance(e, ccxt.NetworkError) else "交易所" if isinstance(e, ccxt.ExchangeError) else "未知"
                self.logger.warning(f"{symbol} : 下单时发生{error_type}错误,正在进行第{retry_count}次重试: {str(e)}")
                if retry_count == max_retries:
                    error_message = f"!! {symbol}: 下单时重试{max_retries}次后仍未成功：{str(e)}"
                    self.logger.error(error_message)
                    raise Exception(error_message)
                time.sleep(1)
      
    def fetch_position(self, symbol):
        """_summary_

        Args:
            symbol (_type_): _description_

        Returns:
            _type_: _description_
        """
        
        max_retries = 3
        retry_count = 0
        
        while retry_count < max_retries:
            try:
                position = self.exchange.fetch_position(symbol=symbol)
                if position :
                    # self.logger.debug(f"{symbol} 有持仓合约数: {position['contracts']}")
                    return position
                return None
            except Exception as e:
                retry_count += 1
                if retry_count == max_retries:
                    error_message = f"!!{symbol} 获取持仓失败(重试{retry_count}次): {str(e)}"
                    self.logger.error(error_message)
                    raise Exception(error_message)
                   
                self.logger.warning(f"{symbol} 检查持仓失败，正在进行第{retry_count}次重试: {str(e)}")
                time.sleep(0.1)  # 重试前等待0.1秒
    
    def fetch_positions(self):
        """_summary_
        Returns:
            _type_: _description_
        """
        max_retries = 3
        retry_count = 0

        while retry_count < max_retries:
            try:
                positions = self.exchange.fetch_positions()
                return positions
            except Exception as e:
                retry_count += 1
                if retry_count == max_retries:
                    error_message = f"!! 获取持仓列表失败(重试{retry_count}次): {str(e)}"
                    self.logger.error(error_message)
                    raise Exception(error_message)

                self.logger.warning(f"获取持仓列表失败，正在进行第{retry_count}次重试: {str(e)}")
                time.sleep(0.1)  # 重试前等待0.1秒
    
    def fetch_open_orders(self,symbol,params={}):
        max_retries = 3
        retry_count = 0
        
        while retry_count < max_retries:
            try:
                orders = self.exchange.fetch_open_orders(symbol=symbol,params=params)
                return orders
            
            except Exception as e:
                retry_count += 1
                if retry_count == max_retries:
                    error_message = f"{symbol} : fetching open orders(retry {retry_count} times): {str(e)}"
                    self.logger.error(error_message)
                    raise Exception(error_message)
                   
                self.logger.warning(f"{symbol} : Error fetching open orders: {str(e)}")
                time.sleep(0.1)  # 重试前等待0.1秒     
    
    def get_market_price(self, symbol) -> Decimal:
        """
        获取最新价格
        Args:
            symbol: 交易对
        """
        max_retries = 3
        retry_count = 0

        while retry_count < max_retries:
            try:
                ticker = self.exchange.fetch_ticker(symbol)
                if ticker and 'last' in ticker:
                    return OPTools.toDecimal(ticker['last'])
                else:
                    raise Exception(f"{symbol} : Unexpected response structure or missing 'last' price")
            except Exception as e:
                retry_count += 1
                if retry_count == max_retries:
                    error_message = f"{symbol} 获取最新价格失败(重试{retry_count}次): {str(e)}"
                    self.logger.error(error_message)
                    raise Exception(error_message)
            
    def get_historical_klines(self, symbol, bar='15m', limit=300, after:str=None, params={}):
        """
        获取历史K线数据
        Args:
            symbol: 交易对
            bar: K线周期
            limit: 数据条数
            after: 之后时间，格式为 "2025-05-21 23:00:00+08:00"
        """
           
        params = {
            **params,
            # 'instId': instId,
        }
        since = None
        if after:
            since = self.exchange.parse8601(after)
            limit = None
            if since:
                params['paginate'] = True
            
        klines = self.exchange.fetch_ohlcv(symbol, timeframe=bar,since=since, limit=limit, params=params)
        # if 'data' in response and len(response['data']) > 0:
        if klines :
            # return response['data']
            return klines
        else:
            raise Exception(f"{symbol} : Unexpected response structure or missing candlestick data")
        
    def get_historical_klines_df(self, symbol, bar='15m', limit=300, after:str=None, params={}) -> pd.DataFrame:
        klines = self.get_historical_klines(symbol, bar=bar, limit=limit, after=after, params=params)
        return self.format_klines(klines)

    def format_klines(self, klines) -> pd.DataFrame:       
        """_summary_
            格式化K线数据
        Args:
            klines (_type_): _description_
        """
        klines_df = pd.DataFrame(klines, columns=['timestamp', 'open', 'high', 'low', 'close', 'volume']) 
        # 转换时间戳为日期时间
        klines_df['timestamp'] = pd.to_datetime(klines_df['timestamp'], unit='ms').dt.tz_localize('UTC').dt.tz_convert('Asia/Shanghai')
  
        return klines_df


import logging
import pandas as pd

from core.smc.SMCStruct import SMCStruct


class SMCFVG(SMCStruct):
    FVG_TOP = "fvg_top"
    FVG_BOT = "fvg_bot"
    FVG_MID = "fvg_mid"
    FVG_SIDE = "fvg_side"
    FVG_WAS_BALANCED = "fvg_was_balanced"

    def __init__(self):
        super().__init__()
        self.logger = logging.getLogger(__name__)

    def find_FVGs(
        self, struct: pd.DataFrame, side, check_balanced=True, start_index=-1
    ) -> pd.DataFrame:
        """_summary_
            寻找公允价值缺口
        Args:
            data (pd.DataFrame): K线数据
            side (_type_): 交易方向 'buy'|'sell'
            threshold (_type_): 阈值价格，通常为溢价和折价区的CE
            check_balanced (bool): 是否检查FVG是否被平衡过,默认为True
            start_index (int): 开始查找索引的起点,默认为-1

        Returns:
            pd.DataFrame: _description_

        """
        # bug2.2.5_1，未到折价区，计算FVG需要前一根K线
        # df = data.copy().iloc[pivot_index:]
        df = (
            struct.copy()
            if start_index == -1
            else struct.copy().iloc[max(0, start_index - 1) :]
        )

        # 检查数据中是否包含必要的列
        check_columns = [self.HIGH_COL, self.LOW_COL]
        self.check_columns(df, check_columns)

        # 处理公允价值缺口
        # 使用向量化操作替代apply，提高性能
        if side == self.BUY_SIDE:
            condition = df[self.HIGH_COL].shift(1) < df[self.LOW_COL].shift(-1)
            side_value = self.BULLISH_TREND
            price_top = df[self.LOW_COL].shift(-1)
            price_bot = df[self.HIGH_COL].shift(1)
        else:
            condition = df[self.LOW_COL].shift(1) > df[self.HIGH_COL].shift(-1)
            side_value = self.BEARISH_TREND
            price_top = df[self.LOW_COL].shift(1)
            price_bot = df[self.HIGH_COL].shift(-1)

        df.loc[:, self.FVG_SIDE] = pd.Series(
            [side_value if x else None for x in condition], index=df.index
        )
        df.loc[:, self.FVG_TOP] = price_top.where(condition, 0)
        df.loc[:, self.FVG_BOT] = price_bot.where(condition, 0)
        df.loc[:, self.FVG_MID] = (df[self.FVG_TOP] + df[self.FVG_BOT]) / 2

        fvg_df = df[
            df[self.FVG_SIDE] == self.BULLISH_TREND
            if side == self.BUY_SIDE
            else df[self.FVG_SIDE] == self.BEARISH_TREND
        ]
        fvg_df = fvg_df.copy()
        if check_balanced:
            # 检查FVG是否被平衡过
            fvg_df.loc[:, self.FVG_WAS_BALANCED] = fvg_df.apply(
                lambda row: any(df.loc[row.name + 2 :, self.LOW_COL] <= row[self.FVG_BOT])
                if side == self.BUY_SIDE
                else any(
                    df.loc[row.name + 2 :, self.HIGH_COL] >= row[self.FVG_TOP]
                ),  
                axis=1,
            )

            fvg_df = fvg_df[~fvg_df[self.FVG_WAS_BALANCED]]

        return fvg_df




import logging
import operator
from core.utils.OPTools import OPTools
from core.smc.SMCBase import SMCBase

class SMCStruct(SMCBase):
    BULLISH_TREND = 'Bullish'
    BEARISH_TREND = 'Bearish'
    STRUCT_COL = "struct"
    STRUCT_HIGH_COL = "struct_high"
    STRUCT_LOW_COL = "struct_low"
    STRUCT_MID_COL = "struct_mid"
    STRUCT_HIGH_INDEX_COL = "struct_high_index"
    STRUCT_LOW_INDEX_COL = "struct_low_index"
    STRUCT_DIRECTION_COL = "struct_direction"
    HIGH_START_COL = "high_start"
    LOW_START_COL = "low_start"
    DIRECTION_COL = "direction"

    def __init__(self):  
        super().__init__()
        self.logger = logging.getLogger(__name__)
      
   
    def build_struct(self, data, tf=None, is_struct_body_break=True):
        """处理价格结构,识别高低点突破和结构方向
        
        Args:
            data: 数据框
            tf: 时间周期
            window: 寻找结构极值的窗口大小
            is_struct_body_break: 是否使用收盘价判断突破
            
        Returns:
            处理后的数据框,包含结构相关列
        """
        # 复制数据并去掉最后一条记录，因为最后一条记录不是完成状态的K线
        # 根据时间周期处理数据框
        df = data.copy()
        if tf == '1m':
            df = df.iloc[:-1]  # 1分钟周期去掉最后一条未完成的K线
            
            
        check_columns = [self.HIGH_COL, self.LOW_COL, self.CLOSE_COL]
        self.check_columns(df, check_columns)
        
        # 初始化结构相关列
        # 定义结构相关的列名
        struct_columns = [self.STRUCT_COL, self.STRUCT_HIGH_COL, self.STRUCT_LOW_COL, 
                         self.STRUCT_HIGH_INDEX_COL, self.STRUCT_LOW_INDEX_COL, self.STRUCT_DIRECTION_COL]
        
        # 初始化结构相关列的默认值
        default_values = {
            self.STRUCT_COL: None,  # 结构类型列初始化为None
            self.STRUCT_HIGH_COL: self.toDecimal('0.0'),  # 结构高点价格初始化为0
            self.STRUCT_LOW_COL: self.toDecimal('0.0'),  # 结构低点价格初始化为0
            self.STRUCT_HIGH_INDEX_COL: 0,  # 结构高点索引初始化为0 
            self.STRUCT_LOW_INDEX_COL: 0,  # 结构低点索引初始化为0
            self.STRUCT_DIRECTION_COL: None  # 结构方向初始化为0
        }
        
        # 为每个结构列赋默认值
        for col in struct_columns:
            df[col] = default_values[col]
            
        # 初始化结构变量
        structure = {
            self.HIGH_COL: df[self.HIGH_COL].iloc[0],
            self.LOW_COL: df[self.LOW_COL].iloc[0],
            self.HIGH_START_COL: -1,
            self.LOW_START_COL: -1,
            self.DIRECTION_COL: 0
        }

        # 确定突破判断列
        break_price_col = self.CLOSE_COL if is_struct_body_break else self.HIGH_COL
        break_price_col_low = self.CLOSE_COL if is_struct_body_break else self.LOW_COL

        for i in range(1, len(df)):
            curr_prices = {
                self.HIGH_COL: self.toDecimal(df[break_price_col].iloc[i]),
                self.LOW_COL: self.toDecimal(df[break_price_col_low].iloc[i])
            }
            
            # 获取前3根K线价格
            prev_prices = {
                self.HIGH_COL: df[break_price_col].iloc[i-3:i].values,
                self.LOW_COL: df[break_price_col_low].iloc[i-3:i].values
            }

            # 判断结构突破
            is_high_broken = self._check_structure_break(
                curr_price=curr_prices[self.HIGH_COL],
                struct_price=structure[self.HIGH_COL],
                prev_prices=prev_prices[self.HIGH_COL],
                struct_start=structure[self.HIGH_START_COL],
                i=i,
                direction=structure[self.DIRECTION_COL],
                target_direction=1,
                mode=self.HIGH_COL
            )
            
            is_low_broken = self._check_structure_break(
                curr_price=curr_prices[self.LOW_COL],
                struct_price=structure[self.LOW_COL],
                prev_prices=prev_prices[self.LOW_COL],
                struct_start=structure[self.LOW_START_COL],
                i=i,
                direction=structure[self.DIRECTION_COL],
                target_direction=2,
                mode=self.LOW_COL
            )
  
            # 处理低点突破
            if is_low_broken:
                extreme_idx = self._get_structure_extreme_bar(
                    df, i, 
                    struct_index=structure[self.LOW_START_COL],
                    mode=self.HIGH_COL
                )
                
                if extreme_idx != 0:
                    struct_type = 'BOS' if structure[self.DIRECTION_COL] == 1 else 'CHOCH'
                    structure = self._handle_structure_break(
                        df, i, structure,
                        break_type=self.LOW_COL,
                        struct_type=struct_type,
                        extreme_idx=extreme_idx
                    )
                else:
                    is_low_broken = False

            # 处理高点突破    
            if is_high_broken:
                extreme_idx = self._get_structure_extreme_bar(
                    df, i,
                    struct_index=structure[self.HIGH_START_COL],
                    mode=self.LOW_COL
                )
                
                if extreme_idx != 0:
                    struct_type = 'BOS' if structure[self.DIRECTION_COL] == 2 else 'CHOCH'
                    structure = self._handle_structure_break(
                        df, i, structure, 
                        break_type=self.HIGH_COL,
                        struct_type=struct_type,
                        extreme_idx=extreme_idx
                    )
                else:
                    is_high_broken = False

            # 如果没有突破发生，更新当前结构
            if not any([is_low_broken, is_high_broken]):
                structure = self._update_current_structure(
                    df, i, structure,
                    is_struct_body_break=is_struct_body_break
                )
            
            # 更新数据框结构列
            self._update_structure_columns(df, i, structure)

        return df

    def _get_structure_extreme_bar(self, df, bar_index,  struct_index, mode='high'):
        """
        获取结构最高点或最低点
        :param df: DataFrame数据
        :param bar_index: 当前K线索引
        :param struct_index: 结构开始索引
        :param mode: high 寻找最高点,low 寻找最低点
        :return: 结构极值点的索引
        """
        df = df.copy()
        # window_start = max(0, bar_index - lookback + 1)
        window_start = max(0, struct_index)
        
        window = df.iloc[window_start : bar_index + 1]
        
        # 获取窗口内的极值点索引
        if mode == self.HIGH_COL:
            # extremeBar = window[self.HIGH_COL].argmax() 
            extremeBar = window[self.HIGH_COL].idxmax() 
            price_col = self.HIGH_COL
            comp_func = lambda x, y: x > y
        else:
            # extremeBar = window[self.LOW_COL].argmin() 
            extremeBar = window[self.LOW_COL].idxmin() 
            price_col = self.LOW_COL
            comp_func = lambda x, y: x < y

        # 初始化记录点
        pivot = 0
        # 从后向前遍历寻找结构极值点
        # for i in range(lookback - 1, -1, -1):
        for idx in range(bar_index - 1, struct_index, -1):
            # 计算当前位置的索引
            # idx = bar_index - i
            if idx - 2 < 0 or idx + 1 >= len(df):
                continue
                
            price_prev = df[price_col].iloc[idx - 1]
            price_prev2 = df[price_col].iloc[idx - 2]
            price_curr = df[price_col].iloc[idx]
            
            # 记录满足条件的点位
            if (comp_func(price_prev, price_prev2) and
                not comp_func(price_curr, price_prev) \
                # and idx - 1 >= extremeBar
                ):
                if pivot == 0:
                    pivot = idx - 1
                    continue
                else:                   
                    # 比较当前点位与之前记录的极值点的价格,因为在区间内有多个极致点，需要比较
                    if comp_func(df[price_col].iloc[idx-1], df[price_col].iloc[pivot]):
                        pivot = idx - 1
                
        return pivot

    def _check_structure_break(self, curr_price, struct_price, prev_prices, struct_start, i, direction, target_direction, mode='high'):
        """检查结构是否突破"""
        comp = operator.gt if mode == self.HIGH_COL else operator.lt
        reverse_comp = operator.le if mode == self.HIGH_COL else operator.ge
        
        basic_break = (
            comp(curr_price, struct_price) and
            all(reverse_comp(p, struct_price) for p in prev_prices) and
            all(i-j > struct_start for j in range(1,4))
        )
        
        direction_break = direction == target_direction and comp(curr_price, struct_price)
        
        return basic_break or direction_break

    def _handle_structure_break(self, df, i, structure, break_type, struct_type, extreme_idx ):
        """处理结构突破"""
        is_high_break = break_type == self.HIGH_COL

        # 更新结构信息
        new_structure = structure.copy()
        new_structure[self.DIRECTION_COL] = 2 if is_high_break else 1
        
        if is_high_break:
            new_structure.update({
                self.LOW_COL: self.toDecimal(df[self.LOW_COL].iloc[extreme_idx]),
                self.HIGH_COL: self.toDecimal(df[self.HIGH_COL].iloc[i]),
                self.LOW_START_COL: extreme_idx,
                self.HIGH_START_COL: i
            })
        else:
            new_structure.update({
                self.HIGH_COL: self.toDecimal(df[self.HIGH_COL].iloc[extreme_idx]),
                self.LOW_COL: self.toDecimal(df[self.LOW_COL].iloc[i]),
                self.HIGH_START_COL: extreme_idx,
                self.LOW_START_COL: i
            })
            
        # 更新DataFrame结构信息
        # df.at[i, self.STRUCT_DIRECTION_COL] = new_structure[self.DIRECTION_COL] 
        df.at[i, self.STRUCT_DIRECTION_COL] = self.BULLISH_TREND if is_high_break else self.BEARISH_TREND
        df.at[i, self.STRUCT_COL] = f"{self.BULLISH_TREND if is_high_break else self.BEARISH_TREND}_{struct_type}"
        
        return new_structure

    def _update_current_structure(self, df, i, structure, is_struct_body_break):
        """更新当前结构"""
        new_structure = structure.copy()
        
        # 检查是否需要更新高点
        if (structure[self.DIRECTION_COL] in (2,0)) and df.at[i,self.HIGH_COL] > structure[self.HIGH_COL]:
            # if not (is_struct_body_break and all(i-j > structure[self.HIGH_START_COL] for j in range(1,4))):
            new_structure[self.HIGH_COL] = df.at[i,self.HIGH_COL]
            new_structure[self.HIGH_START_COL] = i
                
        # 检查是否需要更新低点
        elif (structure[self.DIRECTION_COL] in (1,0)) and df.at[i,self.LOW_COL] < structure[self.LOW_COL]:
            # if not (is_struct_body_break and all(i-j > structure[self.LOW_START_COL] for j in range(1,4))):
            new_structure[self.LOW_COL] = df.at[i,self.LOW_COL]
            new_structure[self.LOW_START_COL] = i
                
        return new_structure

    def _update_structure_columns(self, df, i, structure):
        """更新数据框中的结构列"""
        df.at[i, self.STRUCT_HIGH_COL] = self.toDecimal(structure[self.HIGH_COL])
        df.at[i, self.STRUCT_LOW_COL] = self.toDecimal(structure[self.LOW_COL] )
        df.at[i, self.STRUCT_HIGH_INDEX_COL] = structure[self.HIGH_START_COL]
        df.at[i, self.STRUCT_LOW_INDEX_COL] = structure[self.LOW_START_COL]

    def get_latest_struct(self, df):
        """
        获取最新的结构
        """
        check_columns = [self.STRUCT_COL]
        # if not self.check_columns(df, check_columns):
        try :
            self.check_columns(df, check_columns)
        except ValueError as e:
            df = self.build_struct(df)
 
        data = df.copy()
       
        # 筛选有效结构且在prd范围内的数据
        last_struct = None
        mask = data[self.STRUCT_COL].notna()
        valid_structs = data[ mask ]
        if not valid_structs.empty:
            # 获取最近的结构
            last_struct = valid_structs.iloc[-1]
            return {
                self.STRUCT_COL: last_struct[self.STRUCT_COL],
                self.STRUCT_HIGH_COL: last_struct[self.STRUCT_HIGH_COL],
                self.STRUCT_LOW_COL: last_struct[self.STRUCT_LOW_COL], 
                self.STRUCT_MID_COL: (last_struct[self.STRUCT_HIGH_COL] + last_struct[self.STRUCT_LOW_COL]) / 2,
                self.STRUCT_HIGH_INDEX_COL: last_struct[self.STRUCT_HIGH_INDEX_COL],
                self.STRUCT_LOW_INDEX_COL: last_struct[self.STRUCT_LOW_INDEX_COL],
                self.STRUCT_DIRECTION_COL: last_struct[self.STRUCT_DIRECTION_COL]
            }

        return last_struct

